/*
 * SPDX-FileCopyrightText: 2023-2023 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 *
 */
#ifndef _FCITX_UI_CLASSIC_COLORHELPER_H_
#define _FCITX_UI_CLASSIC_COLORHELPER_H_

#include <utility>
#include "fcitx-utils/color.h"

namespace fcitx::classicui {

std::pair<Color, Color> accentForeground(const Color &accent);
}

#endif
